@@ -34,7 +34,7 @@
 #include FT_FREETYPE_H
 #include FT_SFNT_NAMES_H
 #include FT_TRUETYPE_IDS_H
-#include FT_TYPE1_TABLES_H
+#include FT_TRUETYPE_TABLES_H
 
 #include "ass_utils.h"
 #include "ass.h"
@@ -60,9 +60,8 @@ struct font_info {
     int n_family;
     int n_fullname;
 
-    int slant;
-    int weight;         // TrueType scale, 100-900
-    int width;
+    FT_Long style_flags;
+    int weight;           // TrueType scale, 100-900
 
     // how to access this face
     char *path;            // absolute path
@@ -78,6 +77,9 @@ struct font_info {
 
     // private data for callbacks
     void *priv;
+
+    // unused if the provider has a check_postscript function
+    bool is_postscript;
 };
 
 struct font_selector {
@@ -114,13 +116,6 @@ struct font_data_ft {
     int idx;
 };
 
-static bool check_postscript_ft(void *data)
-{
-    FontDataFT *fd = (FontDataFT *)data;
-    PS_FontInfoRec postscript_info;
-    return !FT_Get_PS_Font_Info(fd->face, &postscript_info);
-}
-
 static bool check_glyph_ft(void *data, uint32_t codepoint)
 {
     FontDataFT *fd = (FontDataFT *)data;
@@ -161,7 +156,6 @@ get_data_embedded(void *data, unsigned char *buf, size_t offset, size_t len)
 
 static ASS_FontProviderFuncs ft_funcs = {
     .get_data          = get_data_embedded,
-    .check_postscript  = check_postscript_ft,
     .check_glyph       = check_glyph_ft,
     .destroy_font      = destroy_font_ft,
 };
@@ -182,7 +176,7 @@ static void load_fonts_from_dir(ASS_Library *library, const char *dir)
             continue;
         ass_msg(library, MSGL_INFO, "Loading font file '%s'", path);
         size_t size = 0;
-        void *data = read_file(library, path, FN_DIR_LIST, &size);
+        void *data = ass_load_file(library, path, FN_DIR_LIST, &size);
         if (data) {
             ass_add_font(library, name, data, size);
             free(data);
@@ -202,6 +196,8 @@ ASS_FontProvider *
 ass_font_provider_new(ASS_FontSelector *selector, ASS_FontProviderFuncs *funcs,
                       void *data)
 {
+    assert(funcs->check_glyph && funcs->destroy_font);
+
     ASS_FontProvider *provider = calloc(1, sizeof(ASS_FontProvider));
     if (provider == NULL)
         return NULL;
@@ -262,7 +258,6 @@ get_font_info(FT_Library lib, FT_Face face, const char *fallback_family_name,
     int num_fullname = 0;
     int num_family   = 0;
     int num_names = FT_Get_Sfnt_Name_Count(face);
-    int slant, weight;
     char *fullnames[MAX_FULLNAME];
     char *families[MAX_FULLNAME];
 
@@ -314,19 +309,15 @@ get_font_info(FT_Library lib, FT_Face face, const char *fallback_family_name,
     if (num_family == 0)
         goto error;
 
-    // calculate sensible slant and weight from style attributes
-    slant  = 110 * !!(face->style_flags & FT_STYLE_FLAG_ITALIC);
-    weight = ass_face_get_weight(face);
-
-    // fill our struct
-    info->slant  = slant;
-    info->weight = weight;
-    info->width  = 100;     // FIXME, should probably query the OS/2 table
+    // calculate sensible weight
+    info->weight = ass_face_get_weight(face);
+    info->style_flags = ass_face_get_style_flags(face);
 
     info->postscript_name = (char *)FT_Get_Postscript_Name(face);
+    info->is_postscript = ass_face_is_postscript(face);
 
     if (num_family) {
-        info->families = calloc(sizeof(char *), num_family);
+        info->families = calloc(num_family, sizeof(char *));
         if (info->families == NULL)
             goto error;
         memcpy(info->families, &families, sizeof(char *) * num_family);
@@ -334,7 +325,7 @@ get_font_info(FT_Library lib, FT_Face face, const char *fallback_family_name,
     }
 
     if (num_fullname) {
-        info->fullnames = calloc(sizeof(char *), num_fullname);
+        info->fullnames = calloc(num_fullname, sizeof(char *));
         if (info->fullnames == NULL)
             goto error;
         memcpy(info->fullnames, &fullnames, sizeof(char *) * num_fullname);
@@ -394,7 +385,6 @@ ass_font_provider_add_font(ASS_FontProvider *provider,
                            int index, void *data)
 {
     int i;
-    int weight, slant, width;
     ASS_FontSelector *selector = provider->parent;
     ASS_FontInfo *info = NULL;
     ASS_FontProviderMetaData implicit_meta = {0};
@@ -450,25 +440,12 @@ ass_font_provider_add_font(ASS_FontProvider *provider,
     for (j = 0; j < meta->n_fullname; j++)
         printf("'%s' ", meta->fullnames[j]);
     printf("\n");
-    printf("  slant: %d\n", meta->slant);
+    printf("  style_flags: %lx\n", meta->style_flags);
     printf("  weight: %d\n", meta->weight);
-    printf("  width: %d\n", meta->width);
     printf("  path: %s\n", path);
     printf("  index: %d\n", index);
 #endif
 
-    weight = meta->weight;
-    slant  = meta->slant;
-    width  = meta->width;
-
-    // check slant/weight for validity, use defaults if they're invalid
-    if (weight < 100 || weight > 900)
-        weight = 400;
-    if (slant < 0 || slant > 110)
-        slant = 0;
-    if (width < 50 || width > 200)
-        width = 100;
-
     // check size
     if (selector->n_font >= selector->alloc_font) {
         selector->alloc_font = FFMAX(1, 2 * selector->alloc_font);
@@ -483,11 +460,11 @@ ass_font_provider_add_font(ASS_FontProvider *provider,
     // set uid
     info->uid = selector->uid++;
 
-    info->slant         = slant;
-    info->weight        = weight;
-    info->width         = width;
+    info->style_flags   = meta->style_flags;
+    info->weight        = meta->weight;
     info->n_fullname    = meta->n_fullname;
     info->n_family      = meta->n_family;
+    info->is_postscript = meta->is_postscript;
 
     info->families = calloc(meta->n_family, sizeof(char *));
     if (info->families == NULL)
@@ -534,6 +511,10 @@ ass_font_provider_add_font(ASS_FontProvider *provider,
     info->provider = provider;
 
     selector->n_font++;
+
+    free_font_info(&implicit_meta);
+    free(implicit_meta.postscript_name);
+
     return true;
 
 error:
@@ -542,9 +523,7 @@ error:
 
     free_font_info(&implicit_meta);
     free(implicit_meta.postscript_name);
-
-    if (provider->funcs.destroy_font)
-        provider->funcs.destroy_font(data);
+    provider->funcs.destroy_font(data);
 
     return false;
 }
@@ -586,10 +565,7 @@ void ass_font_provider_free(ASS_FontProvider *provider)
 
         if (info->provider == provider) {
             ass_font_provider_free_fontinfo(info);
-
-            if (info->provider->funcs.destroy_font)
-                info->provider->funcs.destroy_font(info->priv);
-
+            info->provider->funcs.destroy_font(info->priv);
             info->provider = NULL;
         }
 
@@ -608,9 +584,12 @@ void ass_font_provider_free(ASS_FontProvider *provider)
 static bool check_postscript(ASS_FontInfo *fi)
 {
     ASS_FontProvider *provider = fi->provider;
-    assert(provider && provider->funcs.check_postscript);
+    assert(provider);
 
-    return provider->funcs.check_postscript(fi->priv);
+    if (provider->funcs.check_postscript)
+        return provider->funcs.check_postscript(fi->priv);
+    else
+        return fi->is_postscript;
 }
 
 /**
@@ -670,12 +649,26 @@ static bool matches_full_or_postscript_name(ASS_FontInfo *f,
  */
 static unsigned font_attributes_similarity(ASS_FontInfo *a, ASS_FontInfo *req)
 {
-    unsigned similarity = 0;
-    similarity += ABS(a->weight - req->weight);
-    similarity += ABS(a->slant - req->slant);
-    similarity += ABS(a->width - req->width);
+    unsigned score = 0;
 
-    return similarity;
+    // Assign score for italics mismatch
+    if ((req->style_flags & FT_STYLE_FLAG_ITALIC) &&
+        !(a->style_flags & FT_STYLE_FLAG_ITALIC))
+        score += 1;
+    else if (!(req->style_flags & FT_STYLE_FLAG_ITALIC) &&
+             (a->style_flags & FT_STYLE_FLAG_ITALIC))
+        score += 4;
+
+    int a_weight = a->weight;
+
+    // Offset effective weight for faux-bold (only if font isn't flagged as bold)
+    if ((req->weight > a->weight + 150) && !(a->style_flags & FT_STYLE_FLAG_BOLD))
+        a_weight += 120;
+
+    // Assign score for weight mismatch
+    score += (73 * ABS(a_weight - req->weight)) / 256;
+
+    return score;
 }
 
 #if 0
@@ -696,7 +689,6 @@ static void font_info_dump(ASS_FontInfo *font_infos, size_t len)
         printf("\n");
         printf("  slant: %d\n", font_infos[i].slant);
         printf("  weight: %d\n", font_infos[i].weight);
-        printf("  width: %d\n", font_infos[i].width);
         printf("  path: %s\n", font_infos[i].path);
         printf("  index: %d\n", font_infos[i].index);
         printf("  score: %d\n", font_infos[i].score);
@@ -728,9 +720,8 @@ find_font(ASS_FontSelector *priv,
         return NULL;
 
     // fill font request
-    req.slant   = italic;
-    req.weight  = bold;
-    req.width   = 100;
+    req.style_flags = (italic ? FT_STYLE_FLAG_ITALIC : 0);
+    req.weight      = bold;
 
     // Match font family name against font list
     unsigned score_min = UINT_MAX;
@@ -891,7 +882,7 @@ static char *select_font(ASS_FontSelector *priv,
  * \return font file path
 */
 char *ass_font_select(ASS_FontSelector *priv,
-                      ASS_Font *font, int *index, char **postscript_name,
+                      const ASS_Font *font, int *index, char **postscript_name,
                       int *uid, ASS_FontStream *data, uint32_t code)
 {
     char *res = 0;
@@ -984,7 +975,7 @@ static void process_fontdata(ASS_FontProvider *priv, int idx)
 
         num_faces = face->num_faces;
 
-        charmap_magic(library, face);
+        ass_charmap_magic(library, face);
 
         memset(&info, 0, sizeof(ASS_FontProviderMetaData));
         if (!get_font_info(selector->ftlibrary, face, NULL, &info)) {
@@ -1092,10 +1083,16 @@ ass_fontselect_init(ASS_Library *library, FT_Library ftlibrary, size_t *num_emfo
     priv->path_default = path ? strdup(path) : NULL;
     priv->index_default = 0;
 
+    if (family && !priv->family_default)
+        goto fail;
+    if (path && !priv->path_default)
+        goto fail;
+
     priv->embedded_provider = ass_embedded_fonts_add_provider(priv, num_emfonts);
 
     if (priv->embedded_provider == NULL) {
         ass_msg(library, MSGL_WARN, "failed to create embedded font provider");
+        goto fail;
     }
 
     if (dfp >= ASS_FONTPROVIDER_AUTODETECT) {
@@ -1118,6 +1115,19 @@ ass_fontselect_init(ASS_Library *library, FT_Library ftlibrary, size_t *num_emfo
     }
 
     return priv;
+
+fail:
+    if (priv->default_provider)
+        ass_font_provider_free(priv->default_provider);
+    if (priv->embedded_provider)
+        ass_font_provider_free(priv->embedded_provider);
+
+    free(priv->family_default);
+    free(priv->path_default);
+
+    free(priv);
+
+    return NULL;
 }
 
 void ass_get_available_font_providers(ASS_Library *priv,
